/**
 *  Copyright (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  Licensed under the GPL, Version 3.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/gpl.html
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * --------------------------------------------------------------------------
 *  版权 (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  本开源软件遵循 GPL 3.0 协议;
 *  如果您不遵循此协议，则不被允许使用此文件。
 *  你可以从下面的地址获取完整的协议文本
 *
 *       http://www.gnu.org/licenses/gpl.html
 */
package org.tinygroup.fileresolver.impl;

import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.tinygroup.commons.tools.ClassPathUtil;
import org.tinygroup.fileresolver.FileProcessor;
import org.tinygroup.fileresolver.FileResolver;
import org.tinygroup.fileresolver.ProcessorCallBack;
import org.tinygroup.logger.LogLevel;
import org.tinygroup.logger.Logger;
import org.tinygroup.logger.LoggerFactory;
import org.tinygroup.order.processor.OrderProcessor;
import org.tinygroup.vfs.FileObject;
import org.tinygroup.vfs.VFS;

/**
 * 
 * 功能说明: 文件搜索器默认实现
 * <p>
 * 系统版本: v1.0<br>
 * 开发人员: renhui <br>
 * 开发时间: 2013-2-21 <br>
 * 功能描述: 写明作用，调用方式，使用场景，以及特殊情况<br>
 */
public class FileResolverImpl implements FileResolver {
	private static final String APPLICATION_BEANS_XML = "/Application.beans.xml";
	private static final int DEFAULT_THREAD_NUM = 1;
	private static final String FILE_PROCESSOR_ORDER = "fileprocessor.order.xml";
	private static Logger logger = LoggerFactory
			.getLogger(FileResolverImpl.class);

	// 是否对classPath进行处理，默认为处理
	private boolean classPathResolve = false;
	private int fileProcessorThreadNum = DEFAULT_THREAD_NUM;
	private boolean isOrder;
	// 需要扫描的额外路径列表
	private Set<String> manualClassPaths = new HashSet<String>();
	// 文件处理器列表，由文件查找器统一管理
	private List<FileProcessor> fileProcessorList = new ArrayList<FileProcessor>();
	// 文件信息
	private Map<FileObject, Long> fileDateMap = new HashMap<FileObject, Long>();
	private Map<String, Pattern> skipPathPatternMap = new HashMap<String, Pattern>();

	private Set<String> allScanningPath = new HashSet<String>();

	public List<FileProcessor> getFileProcessorList() {
		return fileProcessorList;
	}

	public void setFileProcessorList(List<FileProcessor> fileProcessorList) {
		this.fileProcessorList = fileProcessorList;
	}

	public FileResolverImpl() {

	}

	public List<String> getManualClassPaths() {
		List<String> classPaths = new ArrayList<String>();
		classPaths.addAll(manualClassPaths);
		return classPaths;
	}

	public void setClassPathResolve(boolean classPathResolve) {
		this.classPathResolve = classPathResolve;
	}

	public void addManualClassPath(String path) {
		if (!manualClassPaths.contains(path)) {
			manualClassPaths.add(path);
		}
	}

	public void addFileProcessor(FileProcessor fileProcessor) {
		fileProcessorList.add(fileProcessor);
	}

	public void resolve() {
		if (fileProcessorList.size() == 0) {
			return;
		} else {
			for (FileProcessor fileProcessor : fileProcessorList) {
				fileProcessor.setFileResolver(this);
			}
			if (isOrder()) {
				OrderProcessor<FileProcessor> processor = new OrderProcessor<FileProcessor>();
				FileObject fileObject = null;
				String jarPath = getClass().getProtectionDomain()
						.getCodeSource().getLocation().getFile();
				fileObject = VFS.resolveFile(jarPath);
				processor.loadOrderFile(fileObject
						.getChild(FILE_PROCESSOR_ORDER));
				fileProcessorList = processor.orderList("fileProcessor",
						fileProcessorList);
			}
			// 移动日志信息，文件搜索器中存在处理器时，才会进行全路径扫描，并打印日志信息
			logger.logMessage(LogLevel.INFO, "正在进行全路径扫描....");
			resolverScanPath();
			for (FileProcessor fileProcessor : fileProcessorList) {
				fileProcessor.process();
			}
			logger.logMessage(LogLevel.INFO, "全路径扫描完成。");
		}
	}

	private void resolverScanPath() {
		cleanProcessor();
		Set<FileObject> classPaths = new HashSet<FileObject>();
		if (classPathResolve) {
			classPaths.addAll(getClassPath());
		}
		classPaths.addAll(resolveManualClassPath());
		try {
			addWebClasses(classPaths);
		} catch (Exception e) {
			logger.errorMessage("查找WEB-INF/classes路径失败！", e);
		}
		try {
			if (classPathResolve) {
				getWebLibJars(classPaths);
			}
		} catch (Exception e) {
			logger.errorMessage("查找Web工程中的jar文件列表失败！", e);
		}
		resolveClassPaths(classPaths);
		resolveDeletedFile();
	}

	void addWebClasses(Set<FileObject> classPaths) {
		logger.logMessage(LogLevel.INFO, "查找WEB-INF/classes路径开始...");
		URL url = FileResolverImpl.class.getResource(APPLICATION_BEANS_XML);
		if (url == null) {
			logger.logMessage(LogLevel.WARN, "{0}不存在，查找WEB-INF/classes路径结束",
					APPLICATION_BEANS_XML);
			return;
		}
		String path = url.toString();
		path = path.substring(0, path.lastIndexOf('/'));
		logger.logMessage(LogLevel.INFO, "WEB-INF/classes路径是:{}", path);
		if (path.indexOf("!") < 0) {// 如果在目录中
			FileObject fileObject = VFS.resolveFile(path);
			classPaths.add(fileObject);
			allScanningPath.add(fileObject.getAbsolutePath());
			String libPath = path.replaceAll("/classes", "/lib");
			logger.logMessage(LogLevel.INFO, "WEB-INF/lib路径是:{}", libPath);
			FileObject libFileObject = VFS.resolveFile(libPath);
			classPaths.add(libFileObject);
			allScanningPath.add(libFileObject.getAbsolutePath());
		} else {// 如果在jar包中
			path = url.getFile().split("!")[0];
			FileObject fileObject = VFS.resolveFile(path);
			classPaths.add(fileObject);
			allScanningPath.add(path);
			String libPath = path.substring(0, path.lastIndexOf('/'));
			logger.logMessage(LogLevel.INFO, "WEB-INF/lib路径是:{}", libPath);
			FileObject libFileObject = VFS.resolveFile(libPath);
			classPaths.add(libFileObject);
			allScanningPath.add(libFileObject.getAbsolutePath());
		}
		logger.logMessage(LogLevel.INFO, "查找WEB-INF/classes路径完成。");
	}

	void getWebLibJars(Set<FileObject> classPaths) throws Exception {
		logger.logMessage(LogLevel.INFO, "查找Web工程中的jar文件列表开始...");
		ClassLoader loader = Thread.currentThread().getContextClassLoader();
		Enumeration<URL> urls = loader.getResources("META-INF/MANIFEST.MF");
		while (urls.hasMoreElements()) {
			URL url = urls.nextElement();
			String path = url.toString();
			path = path.replaceAll("/./", "/");// weblogic中存在这种情况
			if (path.indexOf("!") > 0) {
				path = path.split("!")[0];
			} else {// 专门为JBOSS vfs开头的处理
				path = path.substring(0,
						path.length() - "META-INF/MANIFEST.MF".length() - 1);
				path = path.substring(path.indexOf(':') + 1);
			}
			logger.logMessage(LogLevel.INFO, "扫描到jar文件<{}>。", path);
			FileObject fileObject = VFS.resolveFile(path);
			if (!isSkip(fileObject)) {
				classPaths.add(fileObject);
				allScanningPath.add(path);
			}
		}
		logger.logMessage(LogLevel.INFO, "查找Web工程中的jar文件列表完成。");
	}

	/**
	 * 文件搜索器重新搜索时，需要清空上次各个文件处理器中处理的文件列表,并清除记录的扫描路径。
	 */
	private void cleanProcessor() {
		for (FileProcessor fileProcessor : fileProcessorList) {
			fileProcessor.clean();
		}
		allScanningPath.clear();
	}

	private void resolveDeletedFile() {
		// 处理删除的文件
		Map<FileObject, Long> tempMap = new HashMap<FileObject, Long>();
		for (FileObject fileObject : fileDateMap.keySet()) {
			if (!fileObject.isExist()) {
				// 文件已经被删除
				for (FileProcessor fileProcessor : fileProcessorList) {
					if (!isSkip(fileObject)) {
						fileProcessor.delete(fileObject);
					}
				}
			} else {
				tempMap.put(fileObject, fileDateMap.get(fileObject));
			}
		}
		fileDateMap = tempMap;
	}

	/**
	 * 判断该文件是否被忽略处理
	 * 
	 * @param fileObject
	 * @return
	 */
	boolean isSkip(FileObject fileObject) {
		for (String patternString : skipPathPatternMap.keySet()) {
			Pattern pattern = skipPathPatternMap.get(patternString);
			Matcher matcher = pattern.matcher(fileObject.getAbsolutePath()
					.replaceAll("\\\\", "/"));
			if (matcher.find()) {
				logger.logMessage(LogLevel.INFO, "文件<{}>由于匹配了忽略正则表达式<{}>而被忽略。",
						fileObject, patternString);
				return true;
			}
		}
		return false;
	}

	/**
	 * 
	 * 把系统变量中定义的classpath路径，转化为FileObject对象，并返回FileObject列表
	 * 
	 * @return
	 */
	private List<FileObject> getClassPath() {
		List<FileObject> classPathFileObjects = new ArrayList<FileObject>();
		String classPathProperty = System.getProperties()
				.get("java.class.path").toString();
		String[] classPaths = classPathProperty.split(";");
		for (String classPath : classPaths) {
			if (classPath.length() > 0) {
				FileObject fileObject = VFS.resolveFile(classPath);
				classPathFileObjects.add(fileObject);
				allScanningPath.add(classPath);
			}
		}

		return classPathFileObjects;
	}

	/**
	 * 
	 * 把额外定义的路径，转化为FileObject
	 * 
	 * @return
	 */
	private List<FileObject> resolveManualClassPath() {
		List<FileObject> classPathFileObjects = new ArrayList<FileObject>();
		for (String manualClassPath : manualClassPaths) {
			FileObject fileObject = VFS.resolveFile(manualClassPath);
			classPathFileObjects.add(fileObject);
			allScanningPath.add(manualClassPath);
		}
		return classPathFileObjects;
	}

	/**
	 * 
	 * 文件搜索器根据路径列表进行搜索
	 * 
	 * @param classPaths
	 */
	private void resolveClassPaths(Set<FileObject> classPaths) {
		List<FileObject> scanPathsList = new ArrayList<FileObject>();
		scanPathsList.addAll(classPaths);
		MultiThreadFileProcessor.mutiProcessor(fileProcessorThreadNum,
				"file-resolver-threads", scanPathsList,
				new ProcessorCallBack() {
					public void callBack(FileObject fileObject) {

						logger.logMessage(LogLevel.INFO, "正在扫描路径[{0}]...",
								fileObject.getAbsolutePath());
						resolveFileObject(fileObject);
						logger.logMessage(LogLevel.INFO, "路径[{0}]扫描完成。",
								fileObject.getAbsolutePath());
					}
				});
	}

	private void resolveFileObject(FileObject fileObject) {

		logger.logMessage(LogLevel.DEBUG, "找到文件：{}", fileObject
				.getAbsolutePath().toString());
		// 如果上级为空，且为跳过
		if (fileObject.getParent() == null && isSkip(fileObject)) {
			return;
		}
		processFile(fileObject);
		if (fileObject.isFolder() && fileObject.getChildren() != null) {
			for (FileObject f : fileObject.getChildren()) {
				resolveFileObject(f);
			}
			return;
		}
	}

	/**
	 * 
	 * 文件存在并且不在忽略处理列表中，再交给各个文件处理器进行处理。
	 * 
	 * @param fileObject
	 */
	private synchronized void processFile(FileObject fileObject) {
		if (fileObject.isExist()) {
			for (FileProcessor fileProcessor : fileProcessorList) {
				if (fileProcessor.isMatch(fileObject)) {
					Long lastModifiedTime = fileDateMap.get(fileObject
							.getAbsolutePath().toString());
					long modifiedTime = fileObject.getLastModifiedTime();
					if (lastModifiedTime == null) {// 说明是第一次发现
						fileDateMap.put(fileObject, modifiedTime);
						addFile(fileObject, fileProcessor);
					} else if (lastModifiedTime.longValue() != modifiedTime) {
						changeFile(fileObject, fileProcessor);
						fileDateMap.put(fileObject, modifiedTime);
					} else {
						noChangeFile(fileObject, fileProcessor);
					}
				}
			}
		}
	}

	private void noChangeFile(FileObject fileObject, FileProcessor fileProcessor) {
		fileProcessor.noChange(fileObject);

	}

	private void changeFile(FileObject fileObject, FileProcessor fileProcessor) {
		fileProcessor.modify(fileObject);
	}

	private void addFile(FileObject fileObject, FileProcessor fileProcessor) {
		fileProcessor.add(fileObject);
	}

	public void addSkipPathPattern(String pattern) {
		skipPathPatternMap.put(pattern, Pattern.compile(pattern));
	}

	public boolean getClassPathResolve() {
		return classPathResolve;
	}

	public List<String> getAllScanningPath() {
		List<String> scanList = new ArrayList<String>();
		scanList.addAll(allScanningPath);
		return scanList;
	}

	public FileObject getClassesPath() {
		return VFS.resolveFile(ClassPathUtil.getClassRootPath());
	}

	public boolean isOrder() {
		return isOrder;
	}

	public void setOrder(boolean isOrder) {
		this.isOrder = isOrder;
	}

	public int getFileProcessorThreadNum() {
		return fileProcessorThreadNum;
	}

	public void setFileProcessorThreadNum(int threadNum) {
		this.fileProcessorThreadNum = threadNum;

	}

	/**
	 * 刷新工作
	 */
	public void refresh() {
		if (fileProcessorList.size() == 0) {
			return;
		} else {
			logger.logMessage(LogLevel.INFO, "正在进行全路径刷新....");
			resolverScanPath();
			for (FileProcessor fileProcessor : fileProcessorList) {
				if (fileProcessor.supportRefresh()) {
					fileProcessor.process();
				}
			}
			logger.logMessage(LogLevel.INFO, "全路径刷新结束....");
		}

	}

}
