/**
 *  Copyright (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  Licensed under the GPL, Version 3.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/gpl.html
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * --------------------------------------------------------------------------
 *  版权 (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  本开源软件遵循 GPL 3.0 协议;
 *  如果您不遵循此协议，则不被允许使用此文件。
 *  你可以从下面的地址获取完整的协议文本
 *
 *       http://www.gnu.org/licenses/gpl.html
 */
package org.tinygroup.fileresolver.impl;

import java.lang.reflect.Array;
import java.util.Collection;

import org.tinygroup.fileresolver.FileResolver;
import org.tinygroup.logger.LogLevel;
import org.tinygroup.vfs.FileObject;
import org.tinygroup.xstream.XStreamFactory;
import org.tinygroup.xstream.config.XStreamAnnotationClass;
import org.tinygroup.xstream.config.XStreamClassAlias;
import org.tinygroup.xstream.config.XStreamClassAliases;
import org.tinygroup.xstream.config.XStreamConfiguration;
import org.tinygroup.xstream.config.XStreamPropertyAlias;
import org.tinygroup.xstream.config.XStreamPropertyImplicit;
import org.tinygroup.xstream.config.XStreamPropertyOmit;

import com.thoughtworks.xstream.XStream;

public class XStreamFileProcessor extends AbstractFileProcessor {
	
	private static final String XSTREAM_FILE_EXTENSION = ".xstream.xml";
	private String xstreamFileExtension = XSTREAM_FILE_EXTENSION;

	public String getXstreamFileExtension() {
		return xstreamFileExtension;
	}

	public void setXstreamFileExtension(String xstreamFileExtension) {
		this.xstreamFileExtension = xstreamFileExtension;
	}

	public boolean isMatch(FileObject fileObject) {
		if (fileObject.getFileName().endsWith(XSTREAM_FILE_EXTENSION)) {
			return true;
		}
		return false;
	}

	private void process(FileObject fileObject, ClassLoader loader) {
		try {
			logger.logMessage(LogLevel.INFO, "找到XStream配置文件[{0}]，并开始加载...",
					fileObject.getAbsolutePath());
			XStream loadXStream = XStreamFactory.getXStream();
			XStreamConfiguration xstreamConfiguration = (XStreamConfiguration) loadXStream
					.fromXML(fileObject.getInputStream());
			XStream xStream = XStreamFactory.getXStream(
					xstreamConfiguration.getPackageName(), loader);
			
			ClassLoader newLoader = loader;
			if(newLoader==null)
				newLoader = this.getClass().getClassLoader();
			loadAnnotationClass(xStream, xstreamConfiguration, newLoader);
			if (xstreamConfiguration.getxStreamClassAliases() != null) {
				processClassAliases(xStream,
						xstreamConfiguration.getxStreamClassAliases(), newLoader);
			}
			logger.logMessage(LogLevel.INFO, "XStream配置文件[{0}]，加载完毕。",
					fileObject.getAbsolutePath());
		} catch (Exception e) {
			logger.errorMessage(
					String.format("processing file <%s>", fileObject.getAbsolutePath()),
					e);
		}
	}

	private void processClassAliases(XStream xStream,
			XStreamClassAliases classAliases, ClassLoader loader)
			throws ClassNotFoundException, InstantiationException,
			IllegalAccessException {
		if (classAliases.getClassAliases() != null) {
			for (XStreamClassAlias classAlias : classAliases.getClassAliases()) {
				Class<?> clazz = loader.loadClass(classAlias.getType());
				xStream.alias(classAlias.getAliasName(), clazz);
				processClassAlias(xStream, classAlias, clazz);
			}
		}
	}

	private void processClassAlias(XStream xStream,
			XStreamClassAlias classAlias, Class<?> clazz)
			throws InstantiationException, IllegalAccessException {
		if (classAlias.getProperAliases() != null) {
			processPropertyAlias(xStream, classAlias, clazz);
		}
		if (classAlias.getPropertyImplicits() != null) {
			processPropertyImplicit(xStream, classAlias, clazz);
		}
		if (classAlias.getPropertyOmits() != null) {
			processPropertyOmit(xStream, classAlias, clazz);
		}
	}

	private void processPropertyOmit(XStream xStream,
			XStreamClassAlias classAlias, Class<?> clazz) {
		for (XStreamPropertyOmit propertyOmit : classAlias.getPropertyOmits()) {
			xStream.omitField(clazz, propertyOmit.getAttributeName());
		}
	}

	private void processPropertyImplicit(XStream xStream,
			XStreamClassAlias classAlias, Class<?> clazz)
			throws InstantiationException, IllegalAccessException {
		for (XStreamPropertyImplicit propertyImplicit : classAlias
				.getPropertyImplicits()) {
			String name = propertyImplicit.getAttributeName();
			Object newInstance = clazz.newInstance();
			if (newInstance instanceof Array) {
				xStream.addImplicitArray(clazz, name);
			} else if (newInstance instanceof Collection) {
				xStream.addImplicitCollection(clazz, name);
			}
		}
	}

	private void processPropertyAlias(XStream xStream,
			XStreamClassAlias classAlias, Class<?> clazz) {
		for (XStreamPropertyAlias propertyAlias : classAlias.getProperAliases()) {
			xStream.aliasAttribute(clazz, propertyAlias.getAttributeName(),
					propertyAlias.getAliasName());
		}
	}

	private void loadAnnotationClass(XStream xStream,
			XStreamConfiguration xstreamConfiguration, ClassLoader loader)
			throws ClassNotFoundException {
		if (xstreamConfiguration.getxStreamAnnotationClasses() != null) {
			for (XStreamAnnotationClass annotationClass : xstreamConfiguration
					.getxStreamAnnotationClasses()) {
				xStream.processAnnotations(loader.loadClass(annotationClass
						.getClassName()));
			}
		}
	}


	public void process(ClassLoader loader) {
		for (FileObject fileObject : fileObjects) {
			process(fileObject, loader);
		}

	}


	public void process() {
		process(this.getClass().getClassLoader());

	}

	public void setFileResolver(FileResolver fileResolver) {
		
	}

}
